//!CompilerOption:AddRef:System.Diagnostics.FileVersionInfo.dll
using System;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Media;
using Clio.Utilities;
using ff14bot;
using ff14bot.AClasses;
using ff14bot.Behavior;
using ff14bot.Helpers;
using ff14bot.Managers;
using ICSharpCode.SharpZipLib.Zip;
using Infralution.Localization.Wpf;
using LlamaLibrary;
using LlamaLibrary.Helpers;
using LlamaLibrary.Logging;
using LlamaLibrary.Memory;
using Newtonsoft.Json.Serialization;
using TreeSharp;
using Action = System.Action;

namespace IslandGathererWPFLoader;
public class IslandGathererWPFLoader : BotBase
{
    private const string ProjectName = "IslandGatherer";
    private const string CompiledAssemblyName = "IslandGathererWPF.dll";
    private const string VersionFileName = "Version.txt";
    private const string VersionUrl = @$"https://islandgatherer.llamaserver.com/{VersionFileName}";
    private const string DataUrl = @$"https://islandgatherer.llamaserver.com/{ProjectName}.zip";
    private const bool Debug = false;

    private static readonly LLogger Log = new("IslandGatherer", Colors.Lime);
    //private static readonly string VersionPath = Path.Combine(LoaderFolderName, VersionFileName);
    private readonly string ProjectAssembly;

    private static string _name;
    private static PulseFlags _pulseFlags;
    private static Action _onButtonPress;
    private static Action _start;
    private static Action _stop;
    private static Action _initialize;
    private static Action? _Pulse;
    private static bool _isAutonomous;
    private static bool _wantButton;
    private static bool _requiresProfile;
    private static string _englishName;
    private static Version _version;
    private static Action? _onShutdown;
    private static bool _updated;
    private static readonly object Locker = new();
    private Action _clear;
    private Func<Composite> _getRoot;

    public IslandGathererWPFLoader()
    {
        LoaderFolderName =  GeneralFunctions.SourceDirectory()?.FullName;
        ProjectAssembly = Path.Combine(LoaderFolderName, CompiledAssemblyName);

        if (!Debug)
        {
            Log.Information("Checking for updates...");
            lock (Locker)
            {
                if (_updated)
                {
                    return;
                }

                _updated = true;
            }

            Task.Run(async () => { await Update(LoaderFolderName); }).Wait();
        }

        try
        {
            Unblock(ProjectAssembly);
        }
        catch (Exception e)
        {
            Log.Error("Failed to unblock");
            Log.Exception(e);
        }

        Load();
    }

    private static string LoaderFolderName { get; set; }
    public override Composite Root => _getRoot();
    public override string Name => _name;
    public override bool IsAutonomous => _isAutonomous;
    public override bool WantButton => _wantButton;
    public override bool RequiresProfile => _requiresProfile;
    public override PulseFlags PulseFlags => _pulseFlags;
    public override string EnglishName => _englishName;
    public override Version Version => _version;
    private Assembly? assembly;

    public override void OnButtonPress()
    {
        _onButtonPress?.Invoke();
    }

    public override void Initialize()
    {
        //Log.Information($"Starting Initialize {_initialize?.Method.Name}");
        _initialize?.Invoke();
    }

    public override void Start()
    {
        _start?.Invoke();
    }

    public override void Stop()
    {
        _stop?.Invoke();
    }

    public override void Pulse()
    {
        _Pulse?.Invoke();
    }

    public override void OnShutdown()
    {
        _onShutdown?.Invoke();
    }

    private void Load()
    {
        Log.Information("Starting constructor");

        RedirectAssembly();

        Log.Information("Redirected assemblies");

        if (!File.Exists(ProjectAssembly))
        {
            Log.Error($"Can't find {ProjectAssembly}");
            return;
        }


        assembly = LoadAssembly(ProjectAssembly);

        if (assembly == null)
        {
            return;
        }

        //AssemblyProxy.AddAssembly("Microsoft.Bcl.AsyncInterfaces", typeof(System.IAsyncDisposable).Assembly);
        Type baseType;
        try
        {
            baseType = assembly.DefinedTypes.FirstOrDefault(i => typeof(ICompiledBotbase).IsAssignableFrom(i));
            //Log.Information($"Base:  {baseType.FullName}");
        }
        catch (ReflectionTypeLoadException ex)
        {
            var sb = new StringBuilder();
            foreach (var exSub in ex.LoaderExceptions)
            {
                sb.AppendLine(exSub.Message);
                var exFileNotFound = exSub as FileNotFoundException;
                if (exFileNotFound != null)
                {
                    if (!string.IsNullOrEmpty(exFileNotFound.FusionLog))
                    {
                        sb.AppendLine("Fusion Log:");
                        sb.AppendLine(exFileNotFound.FusionLog);
                    }
                }

                sb.AppendLine();
            }

            var errorMessage = sb.ToString();
            Log.Error(errorMessage);
            return;
            //Display or log the error based on your application.
        }
        catch (Exception e)
        {
            Log.Exception(e);
            return;
        }

        if (baseType == null)
        {
            Log.Error("Base type is null");
            return;
        }

        //dispatcher.BeginInvoke(new Action(() =>
        //{
        ICompiledBotbase compiledAsyncBotbase;
        try
        {
            compiledAsyncBotbase = (ICompiledBotbase)Activator.CreateInstance(baseType);
        }
        catch (ReflectionTypeLoadException ex)
        {
            var sb = new StringBuilder();
            foreach (var exSub in ex.LoaderExceptions)
            {
                sb.AppendLine(exSub.Message);
                var exFileNotFound = exSub as FileNotFoundException;
                if (exFileNotFound != null)
                {
                    if (!string.IsNullOrEmpty(exFileNotFound.FusionLog))
                    {
                        sb.AppendLine("Fusion Log:");
                        sb.AppendLine(exFileNotFound.FusionLog);
                    }
                }

                sb.AppendLine();
            }

            var errorMessage = sb.ToString();
            Log.Error(errorMessage);
            return;
            //Display or log the error based on your application.
        }
        catch (Exception e)
        {
            Log.Exception(e);
            return;
        }

        _name = compiledAsyncBotbase.Name;
        _start = compiledAsyncBotbase.Start;
        _stop = compiledAsyncBotbase.Stop;
        _onButtonPress = compiledAsyncBotbase.OnButtonPress;
        _pulseFlags = compiledAsyncBotbase.PulseFlags;
        _wantButton = compiledAsyncBotbase.WantButton;
        _requiresProfile = compiledAsyncBotbase.RequiresProfile;
        _isAutonomous = compiledAsyncBotbase.IsAutonomous;
        _getRoot = compiledAsyncBotbase.GetRoot;
        _initialize = compiledAsyncBotbase.Initialize;
        _Pulse = baseType.GetMethod("Pulse")?.CreateDelegate(typeof(Action), compiledAsyncBotbase) as Action;
        _englishName = typeof(BotBase).GetProperty("EnglishName")?.GetValue(compiledAsyncBotbase) as string;
        _onShutdown = baseType.GetMethod("OnShutdown")?.CreateDelegate(typeof(Action), compiledAsyncBotbase) as Action;
        _version =  typeof(BotBase).GetProperty("Version")?.GetValue(compiledAsyncBotbase) as Version;
    }

    private static Assembly? LoadAssembly(string path)
    {
        if (!File.Exists(path))
        {
            return null;
        }

        Assembly? assembly = null;
        try
        {
            assembly = Assembly.LoadFrom(path);
        }
        catch (Exception e)
        {
            Logging.WriteException(e);
        }

        return assembly;
    }

    private void RedirectAssembly()
    {
        AssemblyProxy.Init();
        AppDomain.CurrentDomain.AppendPrivatePath(LoaderFolderName);
        AssemblyProxy.AddAssembly("Clio.Localization", typeof(CultureManager).Assembly);
    }

    private async Task Update(string loaderFolderName)
    {
        var local = GetLocalVersion();
        var data = await TryUpdate(local);
        if (data == null)
        {
            return;
        }

        try
        {
            Clean(loaderFolderName);
        }
        catch (Exception e)
        {
            Log.Exception(e);
        }

        try
        {
            Extract(data, loaderFolderName);
        }
        catch (Exception e)
        {
            Log.Exception(e);
        }
    }

    private string GetLocalVersion()
    {
        if (!File.Exists(ProjectAssembly))
        {
            return null;
        }

        try
        {
            //Get file version of assembly
            var fileVersionInfo = FileVersionInfo.GetVersionInfo(ProjectAssembly);

            return GetNumbers(fileVersionInfo.FileVersion);
        }
        catch
        {
            return null;
        }
    }

    private static void Clean(string directory)
    {
        foreach (var file in new DirectoryInfo(directory).GetFiles())
        {
            file.Delete();
        }

        foreach (var dir in new DirectoryInfo(directory).GetDirectories())
        {
            dir.Delete(true);
        }
    }

    private static async Task<byte[]> TryUpdate(string localVersion)
    {
        try
        {
            using var client = new HttpClient();
            var stopwatch = Stopwatch.StartNew();
            var version = GetNumbers(await client.GetStringAsync(VersionUrl));
            Log.Information($"Local: {localVersion} | Latest: {version}");

            if (string.IsNullOrEmpty(version) || version.Equals(localVersion))
            {
                return null;
            }

            Log.Information($"{ProjectName} Updating....");
            using var response = await client.GetAsync(DataUrl);
            if (!response.IsSuccessStatusCode)
            {
                Log.Information($"[Error] Could not download {ProjectName}: {response.StatusCode}");
                return null;
            }

            using var inputStream = await response.Content.ReadAsStreamAsync();
            using var memoryStream = new MemoryStream();
            await inputStream.CopyToAsync(memoryStream);

            stopwatch.Stop();
            Log.Information($"Download took {stopwatch.ElapsedMilliseconds} ms.");

            return memoryStream.ToArray();
        }
        catch (Exception e)
        {
            Log.Information($"[Error] {e}");
            Log.Information($"Local: {localVersion}");
            return null;
        }
    }

    private static string GetNumbers(string input)
    {
        return new string(input.Where(c => char.IsDigit(c) || c == '.').ToArray());
    }

    private static void Extract(byte[] files, string directory)
    {
        using (var stream = new MemoryStream(files))
        {
            var zip = new FastZip();
            zip.ExtractZip(stream, directory, FastZip.Overwrite.Always, null, null, null, false, true);
        }
    }

    [DllImport("kernel32.dll", SetLastError = true)]
    [return: MarshalAs(UnmanagedType.Bool)]
    private static extern bool DeleteFile(string name);

    private static bool Unblock(string fileName)
    {
        return DeleteFile(fileName + ":Zone.Identifier");
    }
}
